/**************************************************************************
 *                                                                        *
 *   Copyright (c) Digital Equipment  Corporation,  1991.   All  Rights   *
 *   Reserved.  Unpublished rights reserved under the copyright laws of   *
 *   the United States.                                                   *
 *                                                                        *
 *   The software  contained  on  this  media  is  proprietary  to  and   *
 *   embodies   the   confidential   technology  of  Digital  Equipment   *
 *   Corporation.  Possession, use, duplication or dissemination of the   *
 *   software  and media is authorized only pursuant to a valid written   *
 *   license from Digital Equipment Corporation.                          *
 *                                                                        *
 *   RESTRICTED RIGHTS LEGEND Use, duplication, or  disclosure  by  the   *
 *   U.S.   Government  is  subject  to  restrictions  as  set forth in   *
 *   Subparagraph  (c)(1)(ii)  of  DFARS  252.227-7013,   or   in   FAR   *
 *   52.227-19, as applicable.                                            *
 *                                                                        *
 **************************************************************************/
#ifdef SCCS
static char sccsid[]="@(#)pick_cube.c	5.3 Stardent 91/02/12";
#endif
/*
			Copyright (c) 1989 by
			Stardent Computer Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Stardent Computer Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under sccs control at Stardent in:
	/sccs3.p/avs/examples/s.pick_cube.c
	
*/
#include <stdio.h>
#include <avs/avs.h>
#include <avs/geom.h>
#include <avs/udata.h>

/*
 * Pick cube example:
 *
 *  This example demonstrates how to use upstream data from the render
 * geometry module to pick geometric objects.  Functionally, this module
 * generates a "cube".  The user selects a face of the cube and the 
 * module regenerates the cube with the selected face highlighted in red.
 * It places a green sphere over the closest vertex to the selection, a 
 * a blue sphere over the selected point.
 */
int pick_cube_desc();

int ((*pick_cube_mod_list[])()) = {
	pick_cube_desc,
};

#define NMODS sizeof(pick_cube_mod_list) / sizeof(char *) 

AVSinit_modules()
{
	AVSinit_from_module_list(pick_cube_mod_list, NMODS);
}

pick_cube_desc()
{
   void pick_cube_init();
   int pick_cube_compute();
   int port;

   AVScreate_output_port("geometry", "geom");
   AVSset_module_name("pick cube", MODULE_DATA);

   AVSadd_parameter("pick down?","boolean",1,0,0);
   AVSadd_parameter("pick moving?","boolean",1,0,0);
   AVSadd_parameter("pick up?","boolean",1,0,0);
   AVSadd_parameter("print matrices?","boolean",0,0,0);
   AVSadd_parameter("select top?","boolean",0,0,0);
   AVSadd_parameter("select children?","boolean",1,0,0);

   port = AVScreate_input_port("pick information", "struct upstream_geom", 
				OPTIONAL | INVISIBLE);
   AVSset_input_class(port, "upstream_geom");

   AVSset_compute_proc(pick_cube_compute);
}

static pick_cube_compute(pick_info,output,down,moving,up,printmat,top,children)
upstream_geom *pick_info;
GEOMedit_list *output;
int down, moving, up;
int printmat, top, children;
{
   char *strrchr();
   int flags;
   int picked_vert, picked_prim;
   float *picked_point;
   *output = GEOMinit_edit_list(*output);

   flags = 0;
   if (down) flags |= BUTTON_DOWN;
   if (moving) flags |= BUTTON_MOVING;
   if (up) flags |= BUTTON_UP;
   
   GEOMedit_selection_mode(*output, "%top",top ? "notify" : "normal", flags);

   /* If the object was picked we can get info about it */
   if (pick_info != NULL && !strncmp(pick_info->picked_obj,"cube",4)) {
      picked_vert = pick_info->vdata;
      picked_prim = pick_info->odata;
      picked_point = &pick_info->mscoord[0];
   }
   else {
      picked_vert = picked_prim = -1;
      picked_point = NULL;
   }

   gen_cube(*output, picked_vert, picked_prim, picked_point);

   GEOMedit_selection_mode(*output,"cube", children ? "notify": "normal",flags);

   if (pick_info != NULL && AVSinput_changed("pick information", 0)) {
      fprintf(stderr,"\n\n\nbutton state is: %s, x = %d, y = %d\n",
	      pick_info->flags == BUTTON_DOWN ? "button down" : 
	      (pick_info->flags == BUTTON_UP ? "button up" : 
	      (pick_info->flags == BUTTON_MOVING ? "button moving" :
	      "not defined")),pick_info->x,pick_info->y);

      fprintf(stderr,"current object name is: %s\n",pick_info->current_obj);
      fprintf(stderr,"mscoord is: %f,%f,%f\n", pick_info->mscoord[0],
					       pick_info->mscoord[1],
					       pick_info->mscoord[2]);
      fprintf(stderr,"wscoord is: %f,%f,%f\n", pick_info->wscoord[0],
					       pick_info->wscoord[1],
					       pick_info->wscoord[2]);
      fprintf(stderr,"sscoord is: %f,%f,%f\n", pick_info->sscoord[0],
					       pick_info->sscoord[1],
					       pick_info->sscoord[2]);

      if (printmat) {
	 fprintf(stderr,"obj xform: \n");
	 print_mat(pick_info->objxform);
	 fprintf(stderr,"world xform: \n");
	 print_mat(pick_info->worldxform);
	 fprintf(stderr,"view xform: \n");
	 print_mat(pick_info->viewxform);
      }

      fprintf(stderr,"camera index is: %d, width = %d, height = %d\n",
	      pick_info->camera_index,pick_info->width,pick_info->height);

      fprintf(stderr,"Picked object is: %s\n",pick_info->picked_obj);

      fprintf(stderr,"Picked vertex is: %f,%f,%f\n",pick_info->vertex[0],
						    pick_info->vertex[1],
						    pick_info->vertex[2]);

      fprintf(stderr,"Picked vertex data is: 0x%x\n",pick_info->vdata);
      fprintf(stderr,"Picked object data is: 0x%x\n",pick_info->odata);
   }

   return(1);
}

print_mat(mat)
float mat[4][4];
{
   int i, j;

   for (i = 0; i < 4; i++) {
      for (j = 0; j < 4; j++) {
	 fprintf(stderr,"%6f ",mat[i][j]);
      }
      fprintf(stderr,"\n");
   }
}

static float Pts[8][3] = {
    -0.75, -0.75, -0.75,  /* 0 */
     0.75, -0.75, -0.75,  /* 1 */
     0.75,  0.75, -0.75,  /* 2 */
    -0.75,  0.75, -0.75,  /* 3 */
    -0.75, -0.75,  0.75,  /* 4 */
     0.75, -0.75,  0.75,  /* 5 */
     0.75,  0.75,  0.75,  /* 6 */
    -0.75,  0.75,  0.75   /* 7 */
};

static int Faces[6][4] = {
    4,3,2,1,
    6,7,8,5,
    2,3,7,6,
    3,4,8,7,
    5,8,4,1,
    2,6,5,1
};

/*
 * gen_cube
 * 
 *  This function generates the data that we are going to try and pick.
 * It creates up to 3 objects: a cube, a red polygon marking one face of   
 * the cube, a green sphere marking the selected vertex, and a blue sphere
 * marking the selected point.
 */
gen_cube(edit_list,vert,prim, pt)
GEOMedit_list edit_list;
int vert, prim; 
float pt[3];
{
   float red[3], green[3], blue[3];
   GEOMobj *obj;
   int i;
   float radii = 0.05;

   obj = GEOMcreate_obj(GEOM_POLYHEDRON,NULL);

   for (i = 0; i < sizeof(Faces)/(sizeof(int)*4); i++) {  
      if (prim != i) {
	 GEOMadd_polygon(obj,4,&Faces[i][0],0,GEOM_COPY_DATA);
	 GEOMadd_prim_data(obj,&i,1,GEOM_COPY_DATA);
      }
   }
   for (i = 0; i < sizeof(Pts)/(sizeof(float)*3); i++) {
      GEOMadd_vertices(obj,&Pts[i][0],1,GEOM_COPY_DATA);
      GEOMadd_vertex_data(obj,&i,1,GEOM_COPY_DATA);
   }

   GEOMgen_normals(obj,GEOM_FACET_NORMALS); 
   
   GEOMedit_geometry(edit_list,"cube",obj);

   GEOMdestroy_obj(obj);

   red[0] = 1.0;
   red[1] = red[2] = 0.0;
   green[1] = 1.0;
   green[0] = green[2] = 0.0;
   blue[2] = 1.0;
   blue[0] = blue[1] = 0.0;

   if (prim != -1) {
      obj = GEOMcreate_obj(GEOM_POLYHEDRON,NULL);
      GEOMadd_polygon(obj,4,&Faces[prim][0],0,GEOM_COPY_DATA);
      for (i = 0; i < sizeof(Pts)/(sizeof(float)*3); i++) {
	 GEOMadd_vertices(obj,&Pts[i][0],1,GEOM_COPY_DATA);
	 GEOMadd_vertex_data(obj,&i,1,GEOM_COPY_DATA);
      }
      GEOMadd_prim_data(obj,&prim,1,GEOM_COPY_DATA);
      GEOMedit_geometry(edit_list,"cube_prim",obj);
      GEOMedit_color(edit_list,"cube_prim",red);
      GEOMedit_parent(edit_list,"cube_prim","cube");

      /*
       * "redirect" selection mode tells the geometry viewer to allow the
       * geometry viewer to pick this object but to report the selections
       * with respect to the parent.  The geometry viewer allows the user to 
       * pick the geometry, but otherwise hides this object from the user.
       */
      GEOMedit_selection_mode(edit_list,"cube_prim","redirect",0); 
      GEOMdestroy_obj(obj);
   }
   /* Delete the highlight object -- nothing was picked of interest */
   else GEOMedit_visibility(edit_list,"cube_prim",-1);


   /* Create a sphere, make it a child of the cube and make it green */
   if (vert != -1) {
      obj = GEOMcreate_sphere(NULL,&Pts[vert][0],&radii,NULL,NULL,1,
				 GEOM_COPY_DATA);
      GEOMedit_geometry(edit_list,"cube_vert",obj);
      GEOMedit_parent(edit_list,"cube_vert","cube");
      GEOMedit_color(edit_list,"cube_vert",green);
      /* 
       * Since this object is just a marker, we tell the geometry viewer
       * to ignore any selections on it.  It is as though the geometry
       * for this object did not exist when the pick occurs (child objects
       * can continue to be picked however). 
       */
      GEOMedit_selection_mode(edit_list,"cube_vert","ignore",0); 
      GEOMdestroy_obj(obj);
   }
   else GEOMedit_visibility(edit_list,"cube_vert",-1);

   /* Now highlight the selected point */
   if (pt != NULL) {
      obj = GEOMcreate_sphere(NULL,pt,&radii,NULL,NULL,1,
				 GEOM_COPY_DATA);
      GEOMedit_geometry(edit_list,"cube_point",obj);
      GEOMedit_parent(edit_list,"cube_point","cube");
      GEOMedit_color(edit_list,"cube_point",blue);
      GEOMedit_selection_mode(edit_list,"cube_point","ignore",0); 
      GEOMdestroy_obj(obj);
   }
   else GEOMedit_visibility(edit_list,"cube_point",-1);
}
